import type { IAdapter } from "../adapters"
import { HMACAssertionError, invariant } from "../errors"
import { OnePasswordFileManager } from "../fs"
import { i18n } from "../i18n"
import type { EncryptedItem } from "./item"
import { Crypto } from "./crypto"
import { Item } from "./item"
import type { Profile, Overview } from "../types"
import { WeakValueMap } from "../weakMap"
import { EventEmitter } from "../ee"
import { asyncMap } from "../util"

type Band = { [uuid: string]: Item }
type FoldersMap = { [uuid: string]: Band }

interface VaultEvents {
  lock: void
}

/**
 * Main OnePassword Vault class
 */
export class Vault extends EventEmitter<VaultEvents> {
  // File system interface
  #profile: Profile
  #folders: FoldersMap

  #items: Item[] = []
  #itemsMap = new WeakValueMap<string, Item>()

  #crypto: Crypto

  private constructor(
    profile: Profile,
    folders: FoldersMap,
    items: Item[],
    crypto: Crypto,
    itemsMap: WeakValueMap<string, Item>
  ) {
    super()
    this.#profile = profile
    this.#folders = folders
    this.#items = items
    this.#crypto = crypto
    this.#itemsMap = itemsMap
  }

  /**
   * Create a new OnePassword Vault instance and read all bands.
   * @internal
   */
  static async of(path: string, profileName = "default", adapter: IAdapter) {
    const crypto = new Crypto(i18n)
    const files = OnePasswordFileManager(adapter.fs, path, profileName)
    const profile = JSON.parse(
      stripText(await files.getProfile(), /^var profile\s*=/, ";")
    )
    const folders = JSON.parse(stripText(await files.getFolders(), "loadFolders(", ");"))

    const itemsMap = new WeakValueMap<string, Item>()
    const bands: Item[] = []

    for (let i = 0; i < 36; i++) {
      const letter = i.toString(36).toUpperCase()
      const source = await files.getBand(letter)
      if (!source) continue
      const object: Record<string, EncryptedItem> = JSON.parse(
        stripText(source, "ld(", ");")
      )
      for (const value of Object.values(object)) {
        const item = new Item(crypto, value)
        bands.push(item)
        itemsMap.set(value.uuid, item)
      }
    }

    const attachments = await files.getAttachments()
    await asyncMap(attachments, async att => {
      const file = itemsMap.get(att.itemUUID)
      invariant(file, `Item ${att.itemUUID} of attachment does not exist`)
      file.addAttachment(await att.getFile())
    })

    return new Vault(profile, folders, bands, crypto, itemsMap)
  }

  readonly overviews = Object.freeze({
    get: (condition: string | ((overview: Overview) => boolean)) => {
      this.#crypto.assertUnlocked()
      if (typeof condition === "string") {
        const title = condition
        condition = overview => overview.title === title
      }
      for (const value of this.#items) {
        if (condition(value.overview)) {
          return value
        }
      }
    },

    values: () => {
      this.#crypto.assertUnlocked()
      return Array.from(this.#items.map(x => x.overview))
    },
  })

  /**
   * Unlock this OnePassword vault.
   * @param masterPassword User provided master password. Only the derived
   * master and overview key will be stored within the class.
   */
  unlock(masterPassword: string) {
    try {
      this.#crypto.unlock(this.#profile, masterPassword)
    } catch (e) {
      if (e instanceof HMACAssertionError) {
        throw new Error(i18n.error.invalidPassword)
      }
      throw e
    }
    return this
  }

  /**
   * Remove derived keys stored within the class instance.
   */
  lock() {
    this.#crypto.lock()
    this.emit("lock")
    return this
  }

  get isLocked() {
    return this.#crypto.locked
  }

  getItem(uuid: string) {
    this.#crypto.assertUnlocked()
    return this.#itemsMap.get(uuid)
  }
}

function stripText(text: string, prefix: string | RegExp, suffix: string | RegExp) {
  if (typeof prefix === "string") {
    if (text.startsWith(prefix)) {
      text = text.slice(prefix.length)
    }
  } else {
    const prefixMatch = text.match(prefix)
    if (prefixMatch) {
      text = text.slice(prefixMatch[0].length)
    }
  }
  if (typeof suffix === "string") {
    if (text.endsWith(suffix)) {
      text = text.slice(0, -suffix.length)
    }
  } else {
    const suffixMatch = text.match(suffix)
    if (suffixMatch) {
      text = text.slice(0, -suffixMatch[0].length)
    }
  }
  return text
}
